#
#    This file is part of Nudnik. <https://github.com/salosh/nudnik.git>
#
#    Nudnik is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Nudnik is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Nudnik.  If not, see <http://www.gnu.org/licenses/>.
#
import time
import sys
import random
if (sys.version_info >= (3, 0)):
    import queue as queue
else:
    import Queue as queue
import threading
import socket
import requests
import grpc

import nudnik
import nudnik.utils as utils

class ParserClient(object):

    def __init__(self, host, port, timeout):
        self.host = host
        self.port = port
        self.timeout = timeout
# TODO gRPC does not honor values bigger than 4194304
#        max_message_size = (100 * 1024 * 1024)
#        options=[('grpc.max_message_length', -1), ('grpc.max_recieve_message_length', -1), ('grpc.max_send_message_length', -1)]
        options=[]
        self.channel = grpc.insecure_channel('{}:{}'.format(self.host, self.port), options=options)

        # bind the client to the server channel
        self.stub = nudnik.entity_pb2_grpc.ParserStub(self.channel)

    def get_response_for_request(self, request):
        return self.stub.parse(request, timeout=self.timeout)

class Stream(threading.Thread):
    def __init__(self, cfg, stream_id, stats):
        threading.Thread.__init__(self)
        self.gtfo = False
        self.event = threading.Event()

        self.cfg = cfg
        self.log = utils.get_logger(cfg.debug)
        self.stream_id = stream_id
        self.stats = stats
        self.queue = queue.Queue()
        self.name = '{}-{}'.format(cfg.name, stream_id)
        self.log.debug('Stream {} initiated'.format(self.name))

    def run(self):
        self.log.debug('Stream {} started, sending {} messages per second'.format(self.name, (self.cfg.rate / float(self.cfg.interval))))

        sequence_id = 0

        active_workers = threading.Semaphore(self.cfg.workers)
        for worker_id in range(0, self.cfg.workers):
            active_workers.acquire()
            thread = MessageSender(self.cfg, self.log, self.stream_id, worker_id, active_workers, self.queue, self.stats)
            thread.daemon = True
            thread.start()

        # Wait for all workers to initialize clients
        for worker_id in range(0, self.cfg.workers):
            active_workers.acquire()

        while not self.gtfo:
            time_start = utils.time_ns()

            for index in range(0, self.cfg.rate):

                message_id = (sequence_id * self.cfg.rate) + index
                if (self.cfg.count > 0) and (message_id >= self.cfg.count):
                    self.exit()
                    return

                if self.cfg.protocol == 'grpc':
                    request = nudnik.entity_pb2.Request(name=self.cfg.name,
                                                        stream_id=self.stream_id,
                                                        sequence_id=sequence_id,
                                                        message_id=message_id,
                                                        ctime=utils.time_ns(),
                                                        load=self.cfg.load_list)
                else:
                    headers = dict()
                    for header in self.cfg.headers:
                        headers.update({str(header[0]): str(header[1])})
                    data = self.cfg.request_format.format(name=self.cfg.name,
                                                          stream_id=self.stream_id,
                                                          sequence_id=sequence_id,
                                                          message_id=message_id,
                                                          ctime=utils.time_ns(),
                                                          load=self.cfg.load_list)

                    req = requests.Request(self.cfg.method, 'http://place_holder', data=data, headers=headers)
                    request = req.prepare()
                    request.name = self.cfg.name
                    request.stream_id=self.stream_id
                    request.sequence_id=sequence_id
                    request.message_id=message_id
                    request.ctime=utils.time_ns()
                    request.load=self.cfg.load_list

                self.queue.put(request)

            if self.cfg.vvv:
                self.log.debug('Active workers/tasks: {}/{}'.format(threading.active_count(), self.queue.qsize()))

            sequence_id += 1

            if self.cfg.chaos > 0 and random.randint(0, self.cfg.cycle_per_hour) <= self.cfg.chaos:
                chaos_exception = utils.ChaosException(self.cfg.chaos_string)
                self.log.fatal(chaos_exception)
                self.exit()
                raise chaos_exception

            elapsed = utils.diff_seconds(time_start, utils.time_ns())
            if elapsed < self.cfg.interval:
                self.event.wait(timeout=(self.cfg.interval - elapsed))

    def exit(self):
        self.gtfo = 1
        self.event.set()

class MessageSender(threading.Thread):

    def __init__(self, cfg, log, stream_id, worker_id, active_workers, queue, stats):
        threading.Thread.__init__(self)
        self.gtfo = False
        self.event = threading.Event()
        self.active_workers = active_workers
        self.cfg = cfg
        self.log = log
        self.client = None
        self.host_address = None
        self.host_resolved_at = 0
        self.session = requests.Session()
        self.queue = queue
        self.stats = stats
        self.worker_id = worker_id
        self.name = '{}-{}-{}'.format(cfg.name, stream_id, worker_id)

    def run(self):
        if self.cfg.protocol == 'grpc':
            self.set_client(True)
        self.active_workers.release()

        if self.cfg.vv:
            self.log.debug('MessageSender {} initiated'.format(self.name))

        while not self.gtfo:
            if self.cfg.protocol == 'grpc':
                self.set_client(False)
            else:
                self.resolv_host(False)

            request = self.queue.get()
            request.worker_id = self.worker_id

            if self.cfg.vvvvv:
                self.log.debug('Handling message_id {}'.format(request.message_id))

            retry_count = 0
            try_count = 1 + self.cfg.retry_count
            send_was_successful = False
            while (not send_was_successful and ((self.cfg.retry_count < 0) or (try_count > 0))):
                request.stime=utils.time_ns()

                meta = self.cfg.meta.format(req=request, node=nudnik.metrics.MetricNode()) if self.cfg.meta is not None else None
                request.meta = utils.get_meta(meta, self.cfg.meta_size)

                if getattr(request, 'load', None) is not None:
                    for load in request.load:
                        utils.generate_load(self.log, load, meta)

                response = None
                if self.cfg.protocol == 'grpc':
                    try:
                        response = self.client.get_response_for_request(request)
                    except grpc._channel._Rendezvous as e:
                        resp = {'status_code': 500}
                        response = nudnik.entity_pb2.Response(**resp)
                        self.log.warn('Reinitializing client due to {}'.format(e))
                        self.set_client(True)
                else:
                    try:
                        request.url = '{}://{}:{}{}'.format(self.cfg.protocol, self.host_address, self.cfg.port, self.cfg.path)
                        response = self.session.send(request)
                        if response.status_code >= 200 and response.status_code < 300:
                           response.status_code = 0
                    except Exception as e:
                        response = None
                        self.log.warn('Resending request due to {}'.format(e))
                        self.resolv_host(True)

                if self.cfg.vvvvv:
                    self.log.debug(response)

                timestamp = utils.time_ns()

                send_was_successful = ( (response is not None) and (response.status_code == 0) and (self.stats.get_fail_ratio() >= self.cfg.fail_ratio))

                if send_was_successful:
                    if self.cfg.vvvvv:
                        self.log.debug('Request was successful')
                    self.stats.add_success()
                    stat = nudnik.stats.Stat(request, response, timestamp)
                    self.stats.append(stat)
                else:
                    self.log.warn('Request was not successful')
                    self.stats.add_failure()
                    try_count -= 1
                    retry_count += 1
                    request.rtime=utils.time_ns()
                    request.rcount = retry_count

            if self.cfg.vv:
                total_rtt = utils.diff_seconds(request.ctime, timestamp) * self.cfg.rate
                if total_rtt > self.cfg.interval:
                    self.log.warn('Predicted total rtt {} for rate {} exceeds interval {}'.format(total_rtt, self.cfg.rate, self.cfg.interval))

    def set_client(self, force):
        resolved_elapsed = utils.diff_seconds(self.host_resolved_at, utils.time_ns())
        if resolved_elapsed < self.cfg.dns_ttl and force is False:
            return

        self.resolv_host(True)
        self.client = None
        index = 0
        while self.client is None:
            try:
                self.client = ParserClient(host_address, self.cfg.port, self.cfg.timeout)
            except Exception as e:
                self.log.warn('Reinitializing client due to {}'.format(e))
                self.event.wait(timeout=((index * 100)/1000))
                index += 1

        if self.cfg.vvv:
            self.log.debug('Client to {} initialized'.format(host_address))

    def resolv_host(self, force):
        resolved_elapsed = utils.diff_seconds(self.host_resolved_at, utils.time_ns())
        if resolved_elapsed < self.cfg.dns_ttl and force is False:
            return

        ipv4=True
        ipv6=False
        if ipv6 is True:
            if ipv4 is True:
                family = 0
            else:
                family = 10
        else:
            family = 2

        addresses = []
        index = 0
        while len(addresses) < 1:
            addresses = socket.getaddrinfo(self.cfg.host, 0, family, 1)
            self.event.wait(timeout=((index * 100)/1000))
            index += 1

        self.host_address = addresses[random.randint(0, (len(addresses) - 1))][-1][0]
        self.host_resolved_at = utils.time_ns()
        if self.cfg.vvv:
            self.log.debug('Host {} resolved as {}'.format(self.cfg.host, self.host_address))

    def exit(self):
        self.gtfo = 1
        self.event.set()




#
#    This file is part of Nudnik. <https://github.com/salosh/nudnik.git>
#
#    Nudnik is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Nudnik is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Nudnik.  If not, see <http://www.gnu.org/licenses/>.
#
import os
import sys
import threading
import psutil
import platform

import nudnik.utils as utils
import nudnik.outputs

class Metrics(threading.Thread):

    def __init__(self, cfg):
        threading.Thread.__init__(self)
        self.gtfo = False
        self.event = threading.Event()
        self.cfg = cfg
        self.log = utils.get_logger(cfg.debug)
        self.workers = list()
        self.node = MetricNode()
        self.log.debug('Metrics thread initiated')

    def run(self):
        self.log.debug('Running {}'.format(self.name))
        mode = 'servermetrics' if self.cfg.server else 'clientmetrics'

        while not self.gtfo:
            time_start = utils.time_ns()

            metric = Metric(self.node)

            if self.cfg.debug:
                for strmetric in _parse_metrics(self.log, mode, [metric], self.cfg.metrics_format_stdout):
                    self.log.debug(strmetric)
            elif 'stdout' in self.cfg.metrics:
                for strmetric in _parse_metrics(self.log, mode, [metric], self.cfg.metrics_format_stdout):
                    self.log.info(strmetric)

            if 'file' in self.cfg.metrics:
                thread = MetricsFileOutput(self.log, self.cfg.metrics_file_path, mode, [metric], self.cfg.metrics_format_file)
                thread.start()
                self.workers.append(thread)
            if 'influxdb' in self.cfg.metrics:
                thread = MetricsInfluxdbOutput(self.log, self.cfg.influxdb_url_metrics, mode, [metric], self.cfg.metrics_format_influxdb)
                thread.start()
                self.workers.append(thread)
            if 'prometheus' in self.cfg.metrics:
                thread = MetricsPrometheusOutput(self.log, self.cfg.prometheus_url_metrics, mode, [metric], self.cfg.metrics_format_prometheus)
                thread.start()
                self.workers.append(thread)

            while len(self.workers) > 0:
                for index, thread in enumerate(self.workers):
                    if thread.is_alive():
                        thread.join(0.25)
                    else:
                        self.workers.pop(index)

            elapsed = utils.diff_seconds(time_start, utils.time_ns())
            if elapsed < self.cfg.metrics_interval:
                self.event.wait(timeout=(self.cfg.metrics_interval - elapsed))

    def exit(self):
        self.gtfo = 1
        self.event.set()

class MetricNode(utils.NudnikObject):

    def __init__(self):
        super(MetricNode, self).__init__()
        self.sysname, self.nodename, self.release, self.version, self.machine = os.uname()
        self.major, self.minor, self.micro, self.releaselevel, self.serial = sys.version_info
        for index, v in enumerate(sys.version.split('\n')):
            setattr(self, 'version_{index}'.format(index=index), v)
        self.dist = ''.join(platform.dist())
        self.distribution = ''.join(platform.linux_distribution())
        self.platform = platform.platform()

class MetricCpu(utils.NudnikObject):

    def __init__(self):
        super(MetricCpu, self).__init__()
        self.count = psutil.cpu_count()
        self.usage = psutil.cpu_percent(interval=None, percpu=False)
        freq = psutil.cpu_freq()
        # None on virtual machines
        if freq is not None:
            freq_dict = freq._asdict()
            for key in freq_dict.keys():
                setattr(self, 'freq_{key}'.format(key=key), freq_dict[key])

        stats = psutil.cpu_stats()._asdict()
        for key in stats.keys():
            setattr(self, 'stats_{key}'.format(key=key), stats[key])

        times = psutil.cpu_times()._asdict()
        for key in times.keys():
            setattr(self, '{key}'.format(key=key), times[key])

        times_percent = psutil.cpu_times_percent(interval=None, percpu=False)._asdict()
        for key in times_percent.keys():
            setattr(self, '{key}_percent'.format(key=key), times_percent[key])

class MetricMemory(utils.NudnikObject):
    def __init__(self):
        super(MetricMemory, self).__init__()

        swap = psutil.swap_memory()._asdict()
        for key in swap.keys():
            setattr(self, 'swap_{key}'.format(key=key), swap[key])

        virtual = psutil.virtual_memory()._asdict()
        for key in virtual.keys():
            setattr(self, '{key}'.format(key=key), virtual[key])

class MetricDisk(utils.NudnikObject):
    def __init__(self):
        super(MetricDisk, self).__init__()

        io = psutil.disk_io_counters(perdisk=True)

        for p in psutil.disk_partitions(all=False):
            dev = p.device.replace('/dev/', '')
            entry = p._asdict()

            usage = psutil.disk_usage(p.mountpoint)._asdict()
            entry.update(usage)
            for key in usage.keys():
                if 'percent' not in key:
                    entry[key + '_h'] = _minus_h(usage[key])

            diskio = io[dev]._asdict()
            entry.update(diskio)
            for key in diskio.keys():
                if '_bytes' in key:
                    entry[key + '_h'] = _minus_h(diskio[key])

            for key in entry:
                setattr(self, '{device}_{key}'.format(device=dev, key=key), entry[key])

class MetricNet(utils.NudnikObject):
    def __init__(self):
        super(MetricNet, self).__init__()
        net = psutil.net_io_counters(pernic=False, nowrap=True)._asdict()
        for key in net.keys():
            setattr(self, '{key}'.format(key=key), net[key])

class Metric(utils.NudnikObject):
    def __init__(self, node):
        super(Metric, self).__init__(timestamp=utils.time_ns())
        self.node = node
        self.cpu = MetricCpu()
        self.mem = MetricMemory()
        self.disk = MetricDisk()
        self.net = MetricNet()

class MetricsFileOutput(nudnik.outputs.FileOutput):

    def parse(self):
        for stat in _parse_metrics(self.log, self.mode, self.data, self.format):
            self.parsed_data.append(stat)

class MetricsInfluxdbOutput(nudnik.outputs.InfluxdbOutput):

    def parse(self):
        for stat in _parse_metrics(self.log, self.mode, self.data, self.format):
            self.parsed_data.append(stat)

class MetricsPrometheusOutput(nudnik.outputs.InfluxdbOutput):

    def parse(self):
        for stat in _parse_metrics(self.log, self.mode, self.data, self.format):
            self.parsed_data.append(stat)

def _parse_metrics(log, mode, metrics, dataformat):
    for metric in metrics:
        try:
            metricstring = dataformat.format(mode=mode,
                                         timestamp=metric.timestamp,
                                         node=metric.node,
                                         cpu=metric.cpu,
                                         mem=metric.mem,
                                         disk=metric.disk,
                                         net=metric.net)
        except Exception as e:
            log.fatal('Fatal error occured while parsing provided format: "{}", {}'.format(dataformat, e))
            break

        yield metricstring

def _minus_h(b):
    symbols = ('K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y', 'S')
    prefix = {}
    for i, s in enumerate(symbols):
        prefix[s] = 1 << (i + 1) * 10
    for s in reversed(symbols):
        if b >= prefix[s]:
            value = float(b) / prefix[s]
            return '%.3f%s' % (value, s)
    return '%.3fB' % (b)





#
#    This file is part of Nudnik. <https://github.com/salosh/nudnik.git>
#
#    Nudnik is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Nudnik is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Nudnik.  If not, see <http://www.gnu.org/licenses/>.
#
import threading

import nudnik
import nudnik.utils as utils

class Load(threading.Thread):
    def __init__(self, cfg):
        threading.Thread.__init__(self)
        self.gtfo = False
        self.event = threading.Event()

        self.cfg = cfg
        self.log = utils.get_logger(cfg.debug)
        self.name = '{}'.format(cfg.name)
        self.log.debug('Load thread {} initiated'.format(self.name))

    def run(self):
        self.log.debug('Load thread {} started'.format(self.name))


        while not self.gtfo:
            time_start = utils.time_ns()

            for load in self.cfg.load_list:
                utils.generate_load(self.log, load, self.cfg.meta)

            elapsed = utils.diff_seconds(time_start, utils.time_ns())
            if elapsed < self.cfg.interval:
                self.event.wait(timeout=(self.cfg.interval - elapsed))

    def exit(self):
        self.gtfo = 1
        self.event.set()




#
#    This file is part of Nudnik. <https://github.com/salosh/nudnik.git>
#
#    Nudnik is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Nudnik is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Nudnik.  If not, see <http://www.gnu.org/licenses/>.
#
import threading
import requests_unixsocket

import nudnik.utils as utils
import nudnik.outputs
import nudnik.metrics

import requests.models

class Stats(threading.Thread):

    def __init__(self, cfg):
        threading.Thread.__init__(self)
        self.gtfo = False
        self.event = threading.Event()
        self.lock = threading.Lock()
        self.stats = list()
        self.successful_requests = 0
        self.failed_requests = 0
        self.cfg = cfg
        self.log = utils.get_logger(cfg.debug)
        self.workers = list()

        self.log.debug('Stats thread initiated')

    def run(self):
        self.log.debug('Running {}'.format(self.name))
        mode = 'serverstats' if self.cfg.server else 'clientstats'

        while not self.gtfo:
            time_start = utils.time_ns()

            current_report = list(self.stats)
            current_report_length = len(current_report)
            if current_report_length > 0:
                if self.cfg.vvv:
                    self.log.debug('Reporting {}/{} items'.format(current_report_length, len(self.stats)))

                if self.cfg.debug:
                    for stat in _parse_stats(self.log, mode, current_report, self.cfg.stats_format_stdout, self.cfg.stats_format_retransmit_stdout):
                        self.log.debug(stat)
                elif 'stdout' in self.cfg.stats:
                    for stat in _parse_stats(self.log, mode, current_report, self.cfg.stats_format_stdout, self.cfg.stats_format_retransmit_stdout):
                        self.log.info(stat)

                if 'file' in self.cfg.stats:
                    thread = FileStats(self.log, self.cfg.stats_file_path, mode, current_report, self.cfg.stats_format_file, self.cfg.stats_format_retransmit_file)
                    thread.start()
                    self.workers.append(thread)
                if 'influxdb' in self.cfg.stats:
                    thread = InfluxdbStats(self.log, self.cfg.influxdb_url_stats, mode, current_report, self.cfg.stats_format_influxdb, self.cfg.stats_format_retransmit_influxdb)
                    thread.start()
                    self.workers.append(thread)
                if 'prometheus' in self.cfg.stats:
                    thread = PrometheusStats(self.log, self.cfg.prometheus_url_stats, mode, current_report, self.cfg.stats_format_prometheus, self.cfg.stats_format_retransmit_prometheus)
                    thread.start()
                    self.workers.append(thread)

                for i in range(0, current_report_length):
                    if self.cfg.vvvvv:
                        self.log.debug('Popping {}/{} items'.format(i, current_report_length))
                    self.stats.pop(0)

                while len(self.workers) > 0:
                    for index, thread in enumerate(self.workers):
                        if thread.is_alive():
                            thread.join(0.25)
                        else:
                            self.workers.pop(index)

            elif self.cfg.vvvv:
                self.log.debug('Nothing to report')

            elapsed = utils.diff_seconds(time_start, utils.time_ns())
            if elapsed < self.cfg.stats_interval:
                self.event.wait(timeout=(self.cfg.stats_interval - elapsed))

    def add_success(self):
        with self.lock:
            self.successful_requests += 1

    def add_failure(self):
        with self.lock:
            self.failed_requests += 1

    def get_fail_ratio(self):
        total = self.failed_requests + self.successful_requests
        try:
            current_fail_ratio = float((self.failed_requests / total) * 100)
        except ZeroDivisionError:
            current_fail_ratio = 100.0

        if self.cfg.vvvvv:
            logformat = 'failed={},success={},current_fail_ratio={},conf_fail_ratio={}'
            self.log.debug(logformat.format(self.failed_requests,
                                            self.successful_requests,
                                            current_fail_ratio,
                                            self.cfg.fail_ratio))
        return current_fail_ratio

    def append(self, stat):
        self.stats.append(stat)

    def exit(self):
        self.gtfo = 1
        self.event.set()

class FileStats(nudnik.outputs.FileOutput):
    def __init__(self, log, path, mode, data, format, retransmit_format):
        super(FileStats, self).__init__(log, path, mode, data, format)
        self.retransmit_format = retransmit_format

    def parse(self):
        for stat in _parse_stats(self.log, self.mode, self.data, self.format, self.retransmit_format):
            self.parsed_data.append(stat)

class InfluxdbStats(nudnik.outputs.InfluxdbOutput):
    def __init__(self, log, url, mode, data, format, retransmit_format):
        super(InfluxdbStats, self).__init__(log, url, mode, data, format)
        self.retransmit_format = retransmit_format

    def parse(self):
        for stat in _parse_stats(self.log, self.mode, self.data, self.format, self.retransmit_format):
            self.parsed_data.append(stat)

class PrometheusStats(nudnik.outputs.PrometheusOutput):
    def __init__(self, log, url, mode, data, format, retransmit_format):
        super(PrometheusStats, self).__init__(log, url, mode, data, format)
        self.retransmit_format = retransmit_format

    def parse(self):
        for stat in _parse_stats(self.log, self.mode, self.data, self.format, self.retransmit_format):
            self.parsed_data.append(stat)

class Stat(utils.NudnikObject):
    def __init__(self, request, response, timestamp):
        super(Stat, self).__init__(timestamp)
        self.request = request
        self.response = response

def _parse_stats(log, mode, stats, format, retransmit_format):
    for stat in stats:
        if getattr(stat.request, 'rcount', None) is None or stat.request.rcount == 0:
            dataformat = format
        else:
            dataformat = retransmit_format

        try:
            if type(stat.request) == requests.models.PreparedRequest:
                statstring = dataformat.format(timestamp_str=str(stat.timestamp),
                                           timestamp=stat.timestamp,
                                           mode=mode,
                                           node=nudnik.metrics.MetricNode(),
                                           req=stat.request,
                                           res=stat.response,
                                           cdelta=utils.diff_nanoseconds(stat.request.ctime, stat.request.stime),
                                           rdelta=utils.diff_nanoseconds(getattr(stat.request, 'rtime', 0), stat.request.stime),
                                           sdelta=utils.diff_nanoseconds(stat.request.stime, getattr(stat.response, 'ctime', 0)),
                                           ldelta=utils.diff_nanoseconds(getattr(stat.response, 'ctime', 0), getattr(stat.response, 'ltime', 0)),
                                           pdelta=utils.diff_nanoseconds(getattr(stat.response, 'ltime', 0), getattr(stat.response, 'stime', 0)),
                                           bdelta=utils.diff_nanoseconds(getattr(stat.response, 'stime', 0), stat.timestamp),
                                           rtt=utils.diff_nanoseconds(stat.request.ctime, stat.timestamp))
            else:
                statstring = dataformat.format(timestamp_str=str(stat.timestamp),
                                           timestamp=stat.timestamp,
                                           mode=mode,
                                           node=nudnik.metrics.MetricNode(),
                                           req=stat.request,
                                           res=stat.response,
                                           cdelta=utils.diff_nanoseconds(stat.request.ctime, stat.request.stime),
                                           rdelta=utils.diff_nanoseconds(stat.request.rtime, stat.request.stime),
                                           sdelta=utils.diff_nanoseconds(stat.request.stime, stat.response.ctime),
                                           ldelta=utils.diff_nanoseconds(stat.response.ctime, stat.response.ltime),
                                           pdelta=utils.diff_nanoseconds(stat.response.ltime, stat.response.stime),
                                           bdelta=utils.diff_nanoseconds(stat.response.stime, stat.timestamp),
                                           rtt=utils.diff_nanoseconds(stat.request.ctime, stat.timestamp))
        except Exception as e:
            log.fatal('Fatal error occured while parsing provided format"{}", {}'.format(dataformat, str(e)))
            break

        yield statstring




#
#    This file is part of Nudnik. <https://github.com/salosh/nudnik.git>
#
#    Nudnik is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Nudnik is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Nudnik.  If not, see <http://www.gnu.org/licenses/>.
#
import threading
import requests_unixsocket

class Output(threading.Thread):
    def __init__(self, log, data, format):
        threading.Thread.__init__(self)
        self.log = log
        self.data = data
        self.parsed_data = []
        self.format = format

    def run(self):
        self.parse()
        self.out()

    def parse(self):
        pass

    def out(self):
        pass

class PrintOutput(Output):

    def out(self):
        print(self.format.format(self.data))

class FileOutput(Output):
    def __init__(self, log, path, mode, data, format):
        super(FileOutput, self).__init__(log, data, format)
        self.mode = mode
        self.path = path

    def out(self):
        self.log.debug('Writing {} items to {}'.format(len(self.parsed_data), self.path))
        with open(self.path, 'a') as statsfile:
            statsfile.write('\n'.join(self.parsed_data))
            statsfile.write('\n')

class InfluxdbOutput(Output):
    def __init__(self, log, url, mode, data, format):
        super(InfluxdbOutput, self).__init__(log, data, format)
        self.mode = mode
        self.url = url

    def out(self):
        self.log.debug('Writing {} items to InfluxDB'.format(len(self.parsed_data)))
        session = requests_unixsocket.Session()
        res = session.post(self.url, data='\n'.join(self.parsed_data))
        session.close()
        if res.status_code != 204:
            self.log.error('InfluxDB response: "{}"'.format(res.text))

class PrometheusOutput(Output):
    def __init__(self, log, url, mode, data, format):
        super(PrometheusOutput, self).__init__(log, data, format)
        self.mode = mode
        self.url = url

    def out(self):
        self.log.debug('Writing {} to Prometheus {}'.format(len(self.parsed_data), self.url))
        session = requests_unixsocket.Session()
        # TODO handle PUT
        for stat in self.parsed_data:
            res = session.post(self.url, stat)
            if res.status_code != 202:
                self.log.error('Response: "{}"'.format(res.text))
            return False

        session.close()
        return True

def create_influxdb_database(protocol, host, database_name):
    # https://docs.influxdata.com/influxdb/v1.7/tools/api/
    # https://docs.influxdata.com/influxdb/v1.7/query_language/database_management/#create-database
    query ='q=CREATE DATABASE "{}"'.format(database_name)
    session = requests_unixsocket.Session()
    res = session.post('{}://{}/query?{}'.format(protocol, host, query))
    session.close()
    if res.status_code == 200:
        print('Response to {}: "{}"'.format(query, res.text))
    else:
        print('Response to {}: "{}"'.format(query, res.text))
        raise RuntimeException(res.text)




#
#    This file is part of Nudnik. <https://github.com/salosh/nudnik.git>
#
#    Nudnik is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Nudnik is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Nudnik.  If not, see <http://www.gnu.org/licenses/>.
#
import logging
import argparse
import yaml
import json
import os
import sys
import threading
import subprocess
import time
import re
from datetime import datetime
import random

if sys.version_info < (3, 0):
    from exceptions import IOError as FileNotFoundError

import nudnik
from nudnik.entity_pb2 import Load
import nudnik.outputs

_BILLION = float(10**9)

DEFAULTS = {
    'config_file': '',
    'protocol': 'grpc',
    'host': '127.0.0.1',
    'port': 5410,
    'path': '/',
    'method': 'GET',
    'headers': [['Content-type', 'application/json']],
    'request_format': '{{"request": "ping", "timestamp": "{ctime}" }}',
    'response_format': '{{"status": 200, "response": "pong", "timestamp": "{ctime}" }}',
    'dns_ttl': 10,
    'server': False,
    'name': 'NAME',
    'name_mismatch_error': None,
    'meta': None,
    'meta_size': (4194304 - 48),
    'workers': os.sysconf('SC_NPROCESSORS_ONLN'),
    'streams': 1,
    'initial_stream_index': 0,
    'interval': 1,
    'rate': 1,
    'timeout': 1,
    'count': 0,
    'chaos': 0,
    'chaos_string': 'In all chaos there is a cosmos, in all disorder a secret order. #Carl_Jung_FTW',
    'load': [],
    'retry_count': -1,
    'fail_ratio': 0,
    'ruok': False,
    'ruok_host': '127.0.0.1',
    'ruok_port': 5310,
    'ruok_path': '/ruok',
    'ruok_headers': [['Content-type', 'application/json']],
    'ruok_response_format': '{{"status": 200, "timestamp": "{date}" }}',
    'metrics': [],
    'metrics_interval': 1,
    'metrics_file_path': './nudnikmetrics.out',
    'metrics_format_stdout': '{timestamp},{node.platform},{cpu.usage},{mem.percent}',
    'metrics_format_file': '{timestamp},{node.platform},{cpu.usage},{mem.percent}',
    'metrics_format_influxdb': 'cpu,hostname={node.nodename} {cpu._csv} {timestamp}\nmem,hostname={node.nodename} {mem._csv} {timestamp}\nnet,hostname={node.nodename} {net._csv} {timestamp}',
    'metrics_format_prometheus': '# TYPE nudnik_metrics summary\nnudnik_metrics{{distro="{node.platform}",cpu_usage="{cpu.usage}", percent="{mem.percent}"}} {timestamp}\n',
    'stats': [],
    'stats_interval': 1,
    'stats_file_path': './nudnikstats.out',
    'stats_format_stdout': '{timestamp_str},{res.status_code},{req.name},{req.message_id},{req.ctime},{cdelta},rtt={rtt}',
    'stats_format_retransmit_stdout': '{timestamp_str},{res.status_code},{req.name},{req.message_id},{req.ctime},{req.rtime},{cdelta},{rdelta},{req.rcount},rtt={rtt}',
    'stats_format_file': '{timestamp_str},{res.status_code},{req.name},{req.message_id},{req.ctime},{cdelta},rtt={rtt}',
    'stats_format_retransmit_file': '{timestamp_str},{res.status_code},{req.name},{req.message_id},{req.ctime},{req.rtime},{cdelta},{rdelta},{req.rcount},rtt={rtt}',
    'stats_format_influxdb': '{mode},hostname={node.nodename},status={res.status_code},name={req.name},sid={req.stream_id},wid={req.worker_id},qid={req.sequence_id} sid={req.stream_id},wid={req.worker_id},mid={req.message_id},ctime={req.ctime},cdelta={cdelta},sdelta={sdelta},pdelta={pdelta},bdelta={bdelta},rtt={rtt} {timestamp}',
    'stats_format_retransmit_influxdb': '{mode},hostname={node.nodename},status={res.status_code},name={req.name},sid={req.stream_id},wid={req.worker_id},qid={req.sequence_id} sid={req.stream_id},wid={req.worker_id},mid={req.message_id},ctime={req.ctime},rtime={req.rtime},cdelta={cdelta},sdelta={sdelta},pdelta={pdelta},bdelta={bdelta},rdelta={rdelta},rcount={req.rcount},rtt={rtt} {timestamp}',
    'stats_format_prometheus': '# TYPE nudnik_message summary\nnudnik_message{{ctime="{req.ctime}",message_id="{req.message_id}", rtt="{rtt}", name="{req.name}"}} {timestamp}\n',
    'stats_format_retransmit_prometheus': '# TYPE nudnik_message summary\nnudnik_message{{ctime="{req.ctime}",message_id="{req.message_id}", rtt="{rtt}", name="{req.name}"}} {timestamp}\n',
    'influxdb_socket_path': '/var/run/influxdb/influxdb.sock',
    'influxdb_protocol': 'http+unix',
    'influxdb_host': '127.0.0.1',
    'influxdb_port': '8086',
    'influxdb_database_prefix': 'nudnik',
    'influxdb_url': '{influxdb_protocol}://{influxdb_host}/write?db={influxdb_database_name}&precision=ns',
    'prometheus_protocol': 'http',
    'prometheus_host': '127.0.0.1',
    'prometheus_port': '9091',
    'prometheus_url': '{prometheus_protocol}://{prometheus_host}:{prometheus_port}/{type}/job/{job_name}/{label_name}/{label_value}',
    'extra': [],
    'debug': False,
    'verbose': 0,
}

class NudnikConfiguration(object):

    def __init__(self):
        self._fields = list()

    def get(self, key, default=None):
        return getattr(self, key, default)

    def set(self, key, value):
        setattr(self, key, value)

    def __iter__(self):
        i = 0
        while i < len(self._fields):
            yield self[i]
            i += 1

    def __len__(self):
        return len(self._fields)

    def __getitem__(self, index):
        return self._fields[index]

    def __setitem__(self, key, value):
        self.set(key, value)

    def __setattr__(self, key, value):
        super(NudnikConfiguration, self).__setattr__(key, value)
        if not key.startswith('_'):
            self._fields.append(key)

    def dict(self):
        d = dict()
        for k in self:
            d.update({k: self.get(k)})
        return d

    def json(self):
        return json.dumps(self.dict())

    def yaml(self):
        return yaml.dump(self.dict())

    def read_config_file(self, file_path, fail_if_missing):
        try:
            with open(file_path, 'r') as ymlfile:
                ymlcfg = yaml.load(ymlfile)

                if not isinstance(ymlcfg, dict):
                    raise FileNotFoundError

                for confkey in DEFAULTS:
                    if confkey in ymlcfg and ymlcfg[confkey] is not None and getattr(self, confkey, None) is None:
                        value = cast_arg_by_type(confkey, ymlcfg[confkey])
                        self.set(confkey, value)

        except yaml.parser.ParserError as e:
            print('Nudnik Configuration error: {}'.format(e))
            sys.exit(1)
        except FileNotFoundError as e:
            print('Unable to read optional configuration file "{}"'.format(file_path))
            if fail_if_missing:
                raise e

def parse_args():
    parser = argparse.ArgumentParser(
        description='Nudnik - gRPC load tester',
        epilog='2019 (C) Salo Shp <https://github.com/salosh/nudnik.git>'
    )
    parser.add_argument('--config-file', '-f',
                        type=str,
                        help='Path to YAML config file')
    parser.add_argument('--host', '-H',
                        type=str,
                        help='host')
    parser.add_argument('--port', '-p',
                        type=int,
                        help='port')
    parser.add_argument('--dns-ttl',
                        type=int,
                        help='Number of seconds before forcing "host" name lookup (Default: 10)')
    parser.add_argument('--server', '-S',
                        action='store_true',
                        default=None,
                        help='Operation mode (default: client)')
    parser.add_argument('--name', '-n',
                        type=str,
                        help='Parser name')
    parser.add_argument('--name-mismatch-error',
                        type=str,
                        choices=['prefix', 'suffix', 'exact'],
                        help='Fail request on name mismatch (default: None)')
    parser.add_argument('--meta', '-M',
                        type=str,
                        help='Send this extra data with every request')
    parser.add_argument('--workers', '-w',
                        type=int,
                        help='Number of workers (Default: Count of CPU cores)')
    parser.add_argument('--streams', '-s',
                        type=int,
                        help='Number of streams (Default: 1)')
    parser.add_argument('--initial-stream-index',
                        type=int,
                        help='Calculate stream ID from this initial index (Default: 0)')
    parser.add_argument('--interval', '-i',
                        type=int,
                        help='Number of seconds per stream message cycle (Default: 1)')
    parser.add_argument('--rate', '-r',
                        type=int,
                        help='Number of messages per interval (Default: 10)')
    parser.add_argument('--timeout',
                        type=int,
                        help='Maximum number of seconds before failing a request (Default: 1)')
    parser.add_argument('--count', '-C',
                        type=int,
                        help='Count of total messages that should be sent (Default: 0 == unlimited)')
    parser.add_argument('--chaos', '-c',
                        type=int,
                        help='Compute statistical process level random crashes [0, 3600/interval] (Default: 0)')
    parser.add_argument('--load', '-l',
                        type=str,
                        nargs=2,
                        action='append',
                        metavar=('load_type', 'load_value'),
                        dest='load',
                        help='Add artificial load [rtt, rttr, cpu, mem, bcmd, fcmd] (Default: None)')
    parser.add_argument('--retry-count',
                        type=int,
                        help='Number of times to re-send failed messages (Default: -1, which means infinite times)')
    parser.add_argument('--fail-ratio',
                        type=str,
                        help='Percent of requests to intentionally fail (Default: 0)')
    parser.add_argument('--ruok', '-R',
                        action='store_true',
                        default=None,
                        help='Enable "Are You OK?" HTTP/1.1 API (default: False)')
    parser.add_argument('--metrics', '-m',
                        type=str,
                        action='append',
                        choices=['stdout', 'file', 'influxdb', 'prometheus'],
                        help='Enable metrics outputs (Default: None)')
    parser.add_argument('--stats', '-t',
                        type=str,
                        action='append',
                        choices=['stdout', 'file', 'influxdb', 'prometheus'],
                        help='Enable stats outputs (Default: None)')
    parser.add_argument('--extra', '-e',
                        type=str,
                        action='append',
                        default=[],
                        help='Extra args (Default: None)')
    parser.add_argument('--debug', '-d',
                        action='store_true',
                        default=None,
                        help='Debug mode (default: False)')
    parser.add_argument('--verbose', '-v',
                        action='count',
                        help='Verbose mode, specify multiple times for extra verbosity (default: None)')
    parser.add_argument('--version', '-V',
                        action='version',
                        version='Nudnik v{} - 2019 (C) Salo Shp <https://github.com/salosh/nudnik.git>'.format(nudnik.__version__),
                        help='Display Nudnik version')

    args = parser.parse_args()
    return args

def parse_config(args):
    cfg = NudnikConfiguration()

    cfg.config_file = DEFAULTS['config_file']

    extra_args = dict()
    for mkv in args.extra:
        arg = mkv.split('=', 1)
        extra_args[arg[0]] = arg[1]

    for key in DEFAULTS:
        env_key_name = 'NUDNIK_{key_name}'.format(key_name=key.upper())
        value = None
        if key in args.__dict__ and vars(args)[key] is not None:
            value = vars(args)[key]
        elif key in extra_args and extra_args[key] is not None:
            value = cast_arg_by_type(key, extra_args[key])
        elif env_key_name in os.environ:
            value = cast_arg_by_type(key, os.environ[env_key_name])

        if value is not None:
            cfg.set(key, value)

    if cfg.config_file != '':
        cfg.read_config_file(cfg.config_file, True)
    else:
        cfg.read_config_file('{}/.nudnikrc'.format(os.environ['HOME']), False)
        cfg.read_config_file('/etc/nudnik/config.yml', False)

    for key in DEFAULTS:
        if cfg.get(key) is None:
            cfg.set(key, DEFAULTS[key])

    if 'influxdb' in cfg.stats or 'influxdb' in cfg.metrics:
        if cfg.influxdb_protocol == 'http+unix':
            cfg.influxdb_host_port = cfg.influxdb_socket_path.replace('/', '%2F')
        else:
            cfg.influxdb_host_port = '{}:{}'.format(cfg.influxdb_host, cfg.influxdb_port)

        database_name_stats = '{}stats'.format(cfg.influxdb_database_prefix)
        database_name_metrics = '{}metrics'.format(cfg.influxdb_database_prefix)

        cfg.influxdb_url_stats = cfg.influxdb_url.format(influxdb_protocol=cfg.influxdb_protocol,
                                               influxdb_host=cfg.influxdb_host_port,
                                               influxdb_database_name=database_name_stats)
        cfg.influxdb_url_metrics = cfg.influxdb_url.format(influxdb_protocol=cfg.influxdb_protocol,
                                               influxdb_host=cfg.influxdb_host_port,
                                               influxdb_database_name=database_name_metrics)

        nudnik.outputs.create_influxdb_database(cfg.influxdb_protocol, cfg.influxdb_host_port, database_name_stats)
        nudnik.outputs.create_influxdb_database(cfg.influxdb_protocol, cfg.influxdb_host_port, database_name_metrics)

    if 'prometheus' in cfg.stats or 'prometheus' in cfg.metrics:
        cfg.prometheus_job_name = 'server' if cfg.server else 'client'
        cfg.prometheus_url_stats = cfg.prometheus_url.format(prometheus_protocol=cfg.prometheus_protocol,
                                                         prometheus_host=cfg.prometheus_host,
                                                         prometheus_port=cfg.prometheus_port,
                                                         type='stats',
                                                         job_name=cfg.prometheus_job_name,
                                                         label_name='instance',
                                                         label_value=os.uname()[1])
        cfg.prometheus_url_metrics = cfg.prometheus_url.format(prometheus_protocol=cfg.prometheus_protocol,
                                                         prometheus_host=cfg.prometheus_host,
                                                         prometheus_port=cfg.prometheus_port,
                                                         type='metrics',
                                                         job_name=cfg.prometheus_job_name,
                                                         label_name='instance',
                                                         label_value=os.uname()[1])

    cfg.cycle_per_hour = int( 3600 / cfg.interval )

    # Clear '%' sign if provided
    fail_ratio = (re.match(r"(([0-9])*(\.)*([0-9])*)", str(cfg.fail_ratio))).groups()[0]
    cfg.fail_ratio = float(fail_ratio) if (fail_ratio != '') else 0.0

    for i in range(1, 6):
        attr = 'v'*i
        cfg.set(attr, cfg.verbose >= i)

    if cfg.v:
        cfg.debug = True

    if cfg.vvvvv:
        print('Python version: {}'.format(sys.version_info))
        print('Effective configuration values:')
        for key in DEFAULTS:
            print('{} - {}'.format(key, getattr(cfg, key)))

    load_list = list()
    if cfg.load is not None:
        for load in cfg.load:
            try:
                load = Load(load_type=load[0], value=load[1])
                load_list.append(load)
            except ValueError as e:
                print('{} - "{}"'.format(e, load[0]))
                sys.exit(1)
    cfg.load_list = load_list
    return cfg

def get_logger(debug=False):
  ''' Initialize logging configuration '''
  format = '%(asctime)-15s %(levelname).5s %(process)d %(threadName)s [%(module)s] [%(funcName)s:%(lineno)d] - %(message)s'
  logging.basicConfig(format=format)
  logger = logging.getLogger(__name__)

  if debug:
      logger.setLevel(logging.DEBUG)
#      logger.debug('Debug-mode logging enabled')
  else:
      logger.setLevel(logging.INFO)

  return logger

def cast_arg_by_type(key, value):
    if type(DEFAULTS[key]) is bool:
        if value in [True, 'TRUE', 'True', 'true', 'YES', 'Yes', 'yes', '1', 1]:
            return True
        elif value in [False, 'FALSE', 'False', 'false', 'NO', 'No', 'no', '0', 0]:
            return False
    elif isinstance(DEFAULTS[key], int):
        return int(value)
    elif isinstance(DEFAULTS[key], str):
        return str(value)

    return value

def generate_load(logger, load, meta):
    if load.load_type == 0:
        time_sleep = float(load.value)
        logger.debug('Sleeping for {}'.format(time_sleep))
        time.sleep(time_sleep)
    elif load.load_type == 1:
        time_sleep = random.uniform(0.0, float(load.value))
        logger.debug('Sleeping for random value {}'.format(time_sleep))
        time.sleep(time_sleep)
    elif load.load_type == 2:
        time_load = load.value
        logger.debug('CPU loading for {} seconds'.format(time_load))
        for i in range(0, os.sysconf('SC_NPROCESSORS_ONLN') * 2):
            load_thread = threading.Thread(target=generate_load_cpu, args=(time_load,));
            load_thread.daemon = True
            load_thread.start()
    elif load.load_type == 3:
        amount_in_mb = load.value
        logger.debug('Loading {} MB to RAM'.format(amount_in_mb))
        load_thread = threading.Thread(target=generate_load_mem, args=(amount_in_mb,));
        load_thread.daemon = True
        load_thread.start()
    elif load.load_type in [4, 5]:
        if load.value == 'meta':
            args = str(meta)
        else:
            args = load.value
        logger.debug('Executing process {}'.format(args.split()))
        load_thread = threading.Thread(target=generate_load_cmd, args=(args,));
        load_thread.daemon = True
        load_thread.start()
        if load.load_type == 5:
            load_thread.join()

def generate_load_cpu(time_load):
    started_at = time.time()
    while ((time.time() - started_at) < float(time_load)):
        pass

def generate_load_mem(amount_in_mb):
    urandom = os.urandom(int(amount_in_mb) * 1024 * 1024)

def generate_load_cmd(args):
    p = subprocess.Popen(args.split(),
                         shell=False,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)

    return p.communicate()

def get_meta(meta, size=0):

    if meta is not None and meta != '':
        if meta[0] == '@':
            meta_filepath = meta[1:]
            if meta_filepath in ['random', 'urandom', '/dev/random', '/dev/urandom']:
                return os.urandom(size)
            with open(meta[1:], 'rb') as f:
                return f.read(size)
        else:
            return meta.encode()

    return ''.encode()

def time_ns():
    # Python < 3.7 doesn't support time.time_ns(), this function unifies stats measurments
    return int(("%.9f" % time.time()).replace('.',''))

def diff_nanoseconds(before, after):
    return (after - before)

def diff_seconds(before, after):
    return ((after - before) / _BILLION)

def time_to_date(timestamp):
    return datetime.fromtimestamp( timestamp / _BILLION )

class ChaosException(Exception): pass

class NudnikObject(object):

    def __init__(self, timestamp=time_ns()):
        self.timestamp = timestamp

    @property
    def _csv(self):
        csv = []
        for key in self.__dict__:
            if not key.startswith('_'):
                csv.append('{k}={v}'.format(k=key, v=getattr(self, key)))
        return ','.join(csv)

    @property
    def _qcsv(self):
        csv = []
        for key in self.__dict__:
            if not key.startswith('_'):
                csv.append('{k}="{v}"'.format(k=key, v=getattr(self, key)))
        return ','.join(csv)
